import random 
import numpy as np 
from tensorflow.keras.optimizers import Adam
import tensorflow as tf 
import math 
import tensorflow.keras as tf 
from sklearn.metrics import accuracy_score
from .errors import *
from .organism import *

class FitnessFunction: 
    def __init__(self): 
        '''
            Params:
                population: Population array from generate_pop method.
                model: Model from build model method.
                train_X, train_y: training data for our model.
                test_X, test_y: testing data for our model.
                scoring_criteria: On what basis scoring, needs to happend eg: Accuracy, f1-score, recall, precision.
        '''
        self.scores = []

    def train_model(self, individual, train_ds):
        """
            This method, will train a neural net model, with the specific individual.
            Params:
                train_X       : Independent variable to the model.
                train_y       : Dependent variable to the model.
                input_dims    : Input dimension of the model, len(individual).
        """
        try: 
            individual.fit(train_ds, epochs=1,
                                   verbose=0, batch_size = 64)
            return individual
        
        except Exception as error:
            return error

    def test_model(self, individual, test_ds, test_X): 
        """
            This method, will return a prediction for the prediction data.
            Params:
                test_X        : Independent Variable of testing data.
                model         : fitted model returned by the train_model method.
                train_X       : Independent Variable of training data.
        """
        try: 
            prediction_score = individual.evaluate(test_ds, 
                                                       verbose=0, batch_size=64)
            pred_y = individual.predict(test_X, batch_size=64, verbose=0)
            return prediction_score, pred_y
        
        except Exception as error:
            return error
        
    def get_prediction_score(self, model, test_ds, test_X, test_y, prediction_score=True): 
        """
            This method, will calculate the metrics value for the prediction done by the model.
            Params: 
                y_true        : Ground Truth value of the testing data.
                y_pred        : Prediction data of the model.
                regression    : To specify, whether it is regression task or classfication task.
        """
        try: 
            if prediction_score: 
                score, y_pred = self.test_model(model, test_ds, test_X)
                test_y = test_y.numpy()
                test_y = [each.argmax() for each in test_y]
                y_pred = [each.argmax() for each in y_pred]
                return accuracy_score(test_y, y_pred)

            results = get_metrics_value(y_true, y_pred, regression)

            if not regression:
                acc, precision, recall, f1_score = results 

            else: rmse, adj_r2 = results

            if self.scoring_criteria == "rmse" and regression: 
                return ((0.9 * rmse) + (0.10 * adj_r2) )

            if self.scoring_criteria == "adj_r2" and regression: 
                return ((0.9 * adj_r2) + (0.1 * rmse))

            if self.scoring_criteria == "acc": 
                return ((0.88 * acc) + (0.03 * recall) + (0.03 * precision) + (0.03 * f1_score))

            if self.scoring_criteria == "recall": 
                return ((0.03 * acc) + (0.88 * recall) + (0.03 * precision) + (0.03 * f1_score))

            if self.scoring_criteria == "precision": 
                return ((0.03 * acc) + (0.03 * recall) + (0.88 * precision) + (0.03 * f1_score))

            if self.scoring_criteria == "f1": 
                return ((0.03 * acc) + (0.03 * recall) + (0.03 * precision) + (0.88 * f1_score))
        
        except Exception as error:
            return error

    def get_fitness_score(self, population, train_ds, test_ds, test_X, test_y): 
        try: 
            fitness_scores = []
            """
                This method calculates the fitness score.
                Params:
                    regression       : To specify, whether it is regression task or classfi
                    cation task.
            """
            for index, individual in enumerate(population): 
                model = individual.model
                trained_model = self.train_model(model, train_ds)
                model_pred_score = self.get_prediction_score(trained_model, test_ds, test_X, test_y, True)
                fitness_scores.append(model_pred_score)
                print(f"      Successfully calculated the fitness score for individual: {index}")

            fitness_scores = np.array(fitness_scores)
            return np.array(fitness_scores / sum(fitness_scores)), fitness_scores
        
        except Exception as error:
            return error 



from .errors import *
from .initial_population_generator import * 
from .next_population_generator import *
from .organism import *
from .fitness_function import *
import tensorflow.keras as keras 
import tensorflow as tf 
import numpy as np 
from tensorflow.keras.layers import (Conv2D, BatchNormalization, MaxPool2D, ReLU,
                                                 ELU, LeakyReLU, Flatten, Dense, AveragePooling2D)

import tensorflow as tf 
import tensorflow.keras as keras 


class GeneticAlgorithm: 
    
    def __init__(self, X_train, y_train, X_test, y_test): 
        assert len(X_train) == len(y_train), "Number of datapoints in X_train and y_train, does'nt match"
        assert len(X_test) == len(y_test), "Number of datapoints in X_test and y_test, does'nt match"
        self.X_train = X_train 
        self.y_train = y_train 
        self.X_test = X_test
        self.y_test = y_test
        self.mutation_rate = 0.1
        self.survival_rate = 0.60
    
    def evolve(self, num_phase: int, num_generation: int, num_population: int, params: dict, verbose=True): 
        """
            - This function contains all the component of Variable Length Genetic Algorithm. It starts generating the 
                initial population, and for that population fitness score will be computed, from that population, some of the 
                individuals will be survived to mate and it will put into the mating pool, from which the next generation population
                will be generation by doing reproduction(mating), the individuals from the mating pool randomly picked mated with each
                other. Some of the parent will be mutated before it is used for the reproduction.
            - Params:
                num_phase        : phase, directly proportional to number of layers(conv2d). 
                num_generation   : how many generation do we need to evolve our initial pop in each phase.
                num_population   : size of the population.
                params         : Dictionary, where key is hyperparameter, value is value of hyperparameter 
        """
        assert type(num_phase) == int, "Phase datatype should be int"
        assert num_population % 2 == 0, "Input value for number of population, should be even."
        assert num_generation > 1, "Number of generation should be greater than 1"
        assert type(num_phase) == int, "Type Error: Int is required"
        assert type(num_generation) == int, "Type Error: Int is required"
        assert type(num_population) == int, "Type Error: Int is required"
        assert type(params) == dict, "Type Error: Dict is required"
        assert type(verbose) == bool, "Type Error: Bool is required"
        assert params.get("phase0", False) != False, "Params dictionary, need a key name as phase0"
        assert params.get("rest_phases", False) != False, "Params dictionary, need a key name as rest_phases"

        try: 
            prev_phase_best = None 
            population = []
            generator = GenPopulation()
            fitness_function = FitnessFunction()

            train_ds, test_ds, test_X, test_y = preprocess_data((self.X_train, self.y_train), (self.X_test, self.y_test))
            for phase in range(num_phase): 
                print(f"phase: {phase}")
                population = generator.generate(
                    params, 
                    num_population,
                    prev_phase_best,
                    phase
                )
                for generation in range(num_generation):         
                    print(f"    Generation: {generation}") if verbose else None

                    fitness_scores, prediction_scores = fitness_function.get_fitness_score(population, train_ds, 
                                                                                                   test_ds, test_X, test_y)
                    print("prediction scores: ", prediction_scores) if verbose else None

                    if len(population) != len(fitness_scores):
                        raise LengthError("Length of Fitness Scores and Population array does'nt match")

                    num_mating_pool_individuals = int(self.survival_rate * num_population)

                    mating_pool_pop  = []
                    mating_pool_fitness_scores = []

                    fs_copy = fitness_scores.copy()
                    pop_copy = population.copy()

                    for _ in range(num_mating_pool_individuals): 
                        index_val = fs_copy.argmax()
                        mating_pool_pop.append(pop_copy[index_val])
                        mating_pool_fitness_scores.append(fs_copy[index_val])
                        np.delete(fs_copy, index_val)
                        np.delete(pop_copy, index_val)

                    del pop_copy
                    del fs_copy
                    if phase == 0: 
                        param = params.get("phase0")
                    else: 
                        param = params.get("rest_phases")

                    mating_pool_fitness_scores = np.array(mating_pool_fitness_scores)
                    next_gen_population = get_next_generation_population(mating_pool_pop, mating_pool_fitness_scores,
                                                                                            self.mutation_rate, param, num_population)
                    population = next_gen_population

                prev_phase_best_index = fitness_scores.argmax()
                prev_phase_best = population[prev_phase_best_index]

            return population, fitness_scores, prev_phase_best, prediction_scores

        except LengthError as error:
            return error 




import random 
import numpy as np
from .errors import *

def mutate(individual, param):
    """
        This function is used to mutate the childrens based randomly.
        Params:
            mutation_rate     : It is probability value, that decides whether to mutate or not.
    """
    try: 
        individual_chromosome = individual.chromosome
        mutation_index = random.randint(0, len(individual_chromosome)-1)
        print(mutation_index > len(individual_chromosome), "mutate")
        if mutation_index > len(individual_chromosome): 
            raise ValueError("chromosome breakage point value is greater than the chromosome dictionary")
            
        for i, key in enumerate(individual_chromosome): 
            if i == mutation_index: 
                individual.chromosome[key] = random.choice(param[key])

        individual.chromosome = individual_chromosome
        return individual
    
    except ValueError as error:
        return error




import random 
import numpy as np 
from .errors import *

def pick_parents(population, fitness_score): 
    """
        This function, will pick two parent chromosomes from the population.
        Params:
            population      : Array of individuals created from GenPopulation Class.
            fitness_score   : Score calculated for each individual in the population by the fitness class.
    """
    try: 
        parent_1, parent_2 = random.choices(population, fitness_score, k=2)
        return parent_1, parent_2
    
    except Exception as error:
        return error
    
    
def reproduce(parent_1, parent_2):
    """
        This function will generate a new childrens by combining two parents gene.
        Params:
            parent_1       : parent_1 array that is picked by the pick_parent function.
            parent_2       : parent_2 array that is picked by the pick_parent function.
    """
    try:
        child1_chromosome, child2_chromosome = {}, {}
        parent_1_chromosome = parent_1.chromosome
        parent_2_chromosome = parent_2.chromosome
        parent_1_chromosome_length = len(parent_1_chromosome)    
        chromosome_breakage_point = random.randint(1, parent_1_chromosome_length-1) 
        
        if chromosome_breakage_point > parent_1_chromosome_length: 
            raise ValueError("chromosome breakage point value is greater than the chromosome dictionary")

        for i, key in enumerate(parent_1_chromosome): 
            if i <= chromosome_breakage_point: 
                child1_chromosome[key] = parent_1_chromosome[key]

        for i, key in enumerate(parent_2_chromosome): 
            if i <= chromosome_breakage_point: 
                child2_chromosome[key] = parent_2_chromosome[key]

        for i, key in enumerate(parent_2_chromosome): 
            if i > chromosome_breakage_point: 
                child1_chromosome[key] = parent_2_chromosome[key]

        for i, key in enumerate(parent_1_chromosome):       
            if i > chromosome_breakage_point: 
                child2_chromosome[key] = parent_1_chromosome[key]
        return child1_chromosome, child2_chromosome
    
    except ValueError as error:
        return error




import numpy as np 
import random 
from tensorflow.keras.layers import Input, Add, Conv2D, BatchNormalization, GlobalAveragePooling2D, Dense
from tensorflow.keras import Model 
import tensorflow.keras as keras 
import tensorflow as tf 
from .errors import *

class Organism: 
    """
        Organism, is a whole component which includes multiple chromosome, in our case it is one chromosome.
        Attributes:
            params            : HyperParameter Dictionary, that will be used to build a chromosome.
            prev_phase_best   : It is the previous phase best organism object. Best One decided based on fitness score.
            phase             : Phase Number.
    """
    def __init__(self, params, prev_phase_best=None, phase=0): 
        self.params = params 
        self.prev_phase_best = prev_phase_best
        self.phase = phase
        self.chromosome = None
        self.model = None

        if phase != 0:
            self.last_model = prev_phase_best.model

    def build_model(self):
        """
            This method, will build a tensorflow model, by using the hyperparameter and layer details of the self.chromosome. It has
            different structure for phase0 and rest of other phases.
        """
        try: 
            keras.backend.clear_session()
            inputs = Input(shape=(32,32,3))
            if self.phase != 0:
                intermediate_model = Model(inputs=self.last_model.input,
                                           outputs=self.last_model.layers[-3].output)
                for layer in intermediate_model.layers:
                    layer.trainable = False
                inter_inputs = intermediate_model(inputs)
                x = Conv2D(filters=self.chromosome['a_output_channels'],
                           padding='same',
                           kernel_size=self.chromosome['a_filter_size'],
                           use_bias=self.chromosome['a_include_BN'])(inter_inputs)

                self.chromosome['activation_type'] = self.prev_phase_best.chromosome['activation_type']
            else:

                x = Conv2D(filters=self.chromosome['a_output_channels'],
                           padding='same',
                           kernel_size=self.chromosome['a_filter_size'],
                           use_bias=self.chromosome['a_include_BN'])(inputs)
            if self.chromosome['a_include_BN']:
                x = BatchNormalization()(x)
            x = self.chromosome['activation_type']()(x)
            if self.chromosome['include_pool']:
                x = self.chromosome['pool_type'](strides=(1,1),
                                                 padding='same')(x)
            if self.phase != 0 and self.chromosome['include_layer'] == False:

                if self.chromosome['include_skip']:
                    y = Conv2D(filters=self.chromosome['a_output_channels'],
                               kernel_size=(1,1),
                               padding='same')(inter_inputs)
                    x = Add()([y,x])
                x = GlobalAveragePooling2D()(x)
                x = Dense(10, activation='softmax')(x)
            else:

                x = Conv2D(filters=self.chromosome['b_output_channels'],
                           padding='same',
                           kernel_size=self.chromosome['b_filter_size'],
                           use_bias=self.chromosome['b_include_BN'])(x)
                if self.chromosome['b_include_BN']:
                    x = BatchNormalization()(x)
                x = self.chromosome['activation_type']()(x)
                if self.chromosome['include_skip']:
                    y = Conv2D(filters=self.chromosome['b_output_channels'],
                               padding='same',
                               kernel_size=(1,1))(inputs)
                    x = Add()([y,x])
                x = GlobalAveragePooling2D()(x)
                x = Dense(10, activation='softmax')(x)
            model = Model(inputs=[inputs], outputs=[x])
            model.compile(optimizer='adam',
                               loss='categorical_crossentropy',
                               metrics=['accuracy']) 
            self.model = model
        
        except Exception as error:
            return error

    def build_chromosome(self): 
        """
            This Method will build a chromosome, by combining a genes(different hyperparameter values). Once it build a chromosome,
            with random hyperparameter values, it will assign the chromosome dictionary to the property(self.chromosome). Phase0, has 
            different chromosome that reset other phases.
        """
        try: 
            if self.phase == 0:
                phase_params = self.params["phase0"]
                self.chromosome = {
                    'a_filter_size': random.choice(phase_params['a_filter_size']),
                    'a_include_BN' : random.choice(phase_params['a_include_BN']),
                    'a_output_channels' : random.choice(phase_params['a_output_channels']),
                    'activation_type' : random.choice(phase_params['activation_type']),
                    'b_filter_size' : random.choice(phase_params['b_filter_size']),
                    'b_include_BN' : random.choice(phase_params['b_include_BN']),
                    'b_output_channels' : random.choice(phase_params['b_output_channels']),
                    'include_pool' : random.choice(phase_params['include_pool']),
                    'pool_type' : random.choice(phase_params['pool_type']),
                    'include_skip' : random.choice(phase_params['include_skip']),
                }

            else: 
                phase_params = self.params["rest_phases"]
                self.chromosome = {
                    'a_filter_size': random.choice(phase_params['a_filter_size']),
                    'a_include_BN' : random.choice(phase_params['a_include_BN']),
                    'a_output_channels' : random.choice(phase_params['a_output_channels']),
                    'include_layer' : random.choice(phase_params['include_layer']),
                    'b_filter_size' : random.choice(phase_params['b_filter_size']),
                    'b_include_BN' : random.choice(phase_params['b_include_BN']),
                    'b_output_channels' : random.choice(phase_params['b_output_channels']),
                    'include_pool' : random.choice(phase_params['include_pool']),
                    'pool_type' : random.choice(phase_params['pool_type']),
                    'include_skip' : random.choice(phase_params['include_skip']),
                }
        
        except Exception as error:
            return error



import random
import numpy as np 
from typing import *
import tensorflow as tf 
from .errors import *
from .organism import *

class GenPopulation:
    def __init__(self): 
        self.population = []

    def generate(self, params: dict, n_population: int, prev_phase_best: Organism, phase_no: int) -> List: 
        """
            Params: 
                number_of_features: is used to encode the actual data into genotype.
                max_pop_size:       is used to restrict number of individual generation.
                max_featrures:      how many features needed to be in the subset. 
                                      if max_features is 0, then maximum subset size is number of number 
        """
        try: 
            self.population = []
            for individual in range(n_population):
                _organism = Organism(
                    params = params,
                    prev_phase_best = prev_phase_best,
                    phase = phase_no
                )
                _organism.build_chromosome()
                _organism.build_model()
                self.population.append(_organism)

            return np.array(self.population)
        
        except Exception as error:
            return error



import tensorflow as tf 
import tensorflow.keras as keras 
import numpy as np
from .errors import *


def preprocess_data(training_data, testing_data): 
    """
        This Function used to convert the tensorflow data object into tensorflow eager object.
        Params: 
            training_data    : training data of the dataset.
            testing_data     : testing data of the dataset.
    """
    try: 
        (X_train, y_train), (X_test, y_test) = training_data, testing_data

        X_train = X_train.astype('float32')
        X_train = X_train/255.

        X_test = X_test.astype('float32')
        X_test = X_test/255.

        y_train = tf.reshape(tf.one_hot(y_train, 10), shape=(-1, 10))
        y_test = tf.reshape(tf.one_hot(y_test, 10), shape=(-1, 10))

        BATCH_SIZE = 256
        AUTOTUNE = tf.data.experimental.AUTOTUNE

        train_ds = tf.data.Dataset.from_tensor_slices((X_train, y_train))
        train_ds = train_ds.shuffle(1024).cache().batch(BATCH_SIZE).prefetch(AUTOTUNE)

        test_ds = tf.data.Dataset.from_tensor_slices((X_test, y_test))
        test_ds = test_ds.cache().batch(BATCH_SIZE).prefetch(AUTOTUNE)

        return train_ds, test_ds, X_test, y_test
    
    except Exception as error:
        return error



from .errors import *    
from .mutation import * 
from .crossover import * 
import random 
import numpy as np 
from .organism import *

def get_next_generation_population(population, fitness_score, mutation_rate, param, n_population): 
    """
        This function used to generate a new individauls by combining genes of the parensts from old population.
        Params:
            population      : Array of individuals created from GenPopulation Class.
            fitness_score   : Score calculated for each individual in the population by the fitness class.
            mutation_rate   : Amount of mutation needed to be done.
            param           : Hyperparameter dictionary, that will be used by the mutate function.
    """
    try: 
        new_generation_population = []
        n_individual_for_next_gen = int(n_population * 0.20)
        fs_copy = fitness_score.copy()
        pop_copy = population.copy()    
        for _ in range(n_individual_for_next_gen): 
            index_val = fs_copy.argmax()
            new_generation_population.append(pop_copy[index_val])
            np.delete(fs_copy, index_val)
            np.delete(pop_copy, index_val)
        
        del fs_copy
        del pop_copy
        
        n_iter = int(n_population//2 - (n_individual_for_next_gen / 2))
        print(n_iter, len(new_generation_population), "n_iter")
        for _ in range(n_iter): 
            parent_1, parent_2 = pick_parents(population, fitness_score)
            # Mutation.
            if random.random() < mutation_rate: 
                parent_1 = mutate(parent_1, param)

            if random.random() < mutation_rate: 
                parent_2 = mutate(parent_2, param)

            # Mating.
            child1_chromosome, child2_chromosome = reproduce(parent_1, parent_2)

            # Creating new organsim for next generation.
            child1_organism = Organism(parent_1.params, parent_1.prev_phase_best, parent_1.phase)
            child1_organism.chromosome = child1_chromosome
            child1_organism.build_model()

            child2_organism = Organism(parent_1.params, parent_1.prev_phase_best, parent_1.phase)
            child2_organism.chromosome = child2_chromosome
            child2_organism.build_model()

            new_generation_population.append(child1_organism)
            new_generation_population.append(child2_organism)
        
        if len(new_generation_population) != n_population: 
            raise LengthError("Length of new_generation_population array is not same as population")
            
        print("[+] Successfully generated the next generation's population.")
        return new_generation_population
    
    except LengthError as error:
        return error



import setuptools
from setuptools import setup

setup()

"""
with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "genetic-algorithm-nas",
    version = "0.1.1",
    author = "Shilash M",
    author_email = "shilu4577@gmail.com",
    description = "Genetic Algorithm for Neural Architecture Search.",

    license = "MIT",
    license_file = "LICENSE",
    long_description = "file:README.md",
    long_description_content_type = "text/markdown",
    project_urls = {
        "Source": "https://github.com/shilu10/Genetic_Algo_NAS",
        
    },
    classifiers = [
       "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
       "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    package_dir = {"": "genetic_algorithm_nas"},
    packages = setuptools.find_packages(where="genetic_algorithm_nas/"),
    python_requires = ">=3.6",
    keywords = [
        "Genetic Algorithm",
        "Neural Architecture Search",
        "HyperParameter Tuning",
    ],
    include_package_data = True,
    zip_safe = True,
    
    test_require = [
        "pytest>=5"
    ],

	
)
"""
